#!/usr/bin/env python
'''
htmlfilter.py
CS 261
'''
import re
from sys import stdin, stdout
from html5lib import HTMLParser

def htmlize(s):
    replacements = [('&', '&amp;'), ('"', '&quot;'), ("'", '&#39;'),
                    ('<', '&lt;'), ('>', '&gt;')]
    for character, replacement in replacements:
        s = s.replace(character, replacement)
    return s

def is_safe_element(elem):
    # From http://wiki.whatwg.org/wiki/Sanitization_rules#Acceptable_Elements
    # with some modifications.
    whitelist = ['a', 'abbr', 'acronym', 'address', 'area', 'b', 'bdo', 'big',
                 'blockquote', 'br', 'button', 'caption', 'center', 'cite',
                 'code', 'col', 'colgroup', 'dd', 'del', 'dfn', 'dir', 'div',
                 'dl', 'dt', 'em', 'fieldset', 'font', 'form', 'h1', 'h2',
                 'h3', 'h4', 'h5', 'h6', 'hr', 'i', 'img', 'input', 'ins',
                 'kbd', 'label', 'legend', 'li', 'map', 'menu', 'ol',
                 'optgroup', 'option', 'p', 'pre', 'q', 's', 'samp', 'select',
                 'small', 'span', 'strike', 'strong', 'sub', 'sup', 'table',
                 'tbody', 'td', 'textarea', 'tfoot', 'th', 'thead', 'tr',
                 'tt', 'u', 'ul', 'var', 'wbr', 'html', 'body', 'plaintext']
    return elem.name.lower() in whitelist

def is_empty_element(elem):
    '''Returns True if the element should not have a closing tag because they
    should have empty content.
    '''
    empty_elements = ['img', 'br', 'wbr', 'hr', 'area', 'col', 'input']
    return not elem.hasContent() and elem.name.lower() in empty_elements

def is_safe_attribute(attr, value):
    '''Returns True if the specified attribute and its value are both allowed.
    For example, even though href is allowed, if its URI does not specify an
    allowed protocol, then the whole attribute is disallowed.
    '''
    # What's important is that the URI begins with a safe protocol.
    uri_pattern = r'(https?|ftp|nntp|xmpp)://[^:]+(:\d+)?[^:]*'
    # We don't have to worry about characters like & or " since we'll
    # encode those entities during the output phase.
    whitelist = {
        'dir'     : '(ltr|rtl)',
        'href'    : uri_pattern,
        'action'  : uri_pattern,
        'src'     : uri_pattern,
        'value'   : '.*',
        'title'   : '.*',
        'alt'     : '.*',
        'rowspan' : r'\d+',
        'colspan' : r'\d+',
        'size'    : r'\d+',
        'tabindex': r'-?\d+',
        'type'    : '[a-z]*',
        'method'  : '(get|put|post|delete)?'
    }
    attr = attr.lower()
    return attr in whitelist and re.match('^' + whitelist[attr] + '$', value, re.I)

# Filter and print out the DOM tree via a depth first search with pre- and
# post-expand subroutines for elements.
def sanitize_node(node):
    '''Filters and returns a node via a depth-first search using pre- and post-
    expand subroutines for elements. This is safe because all data must pass a
    whitelist, and after that must go through HTML entity escaping.
    '''
    sanitized = ''
    if node.type == 1 or node.type == 2:
        # html5lib arbitrarily sets Documents as 1 and DocumentFragments as 2.
        for child in node.childNodes:
            sanitized += sanitize_node(child)
    elif node.type == 4:    # TextNodes are type 4.
        sanitized += htmlize(node.value)
    elif node.type == 5:    # Elements are type 5.
        sanitized += sanitize_element(node)
    # We ignore types 3 (DOCTYPEs) and 6 (CommentNodes).
    return sanitized

def sanitize_element(elem):
    sanitized = ''
    if is_safe_element(elem):
        # Print the opening tag with safe attributes.
        sanitized += '<' + htmlize(elem.name.lower())
        for attr, value in elem.attributes.items():
            if is_safe_attribute(attr, value):
                sanitized += ' %s="%s"' % (htmlize(attr.lower()), htmlize(value))
        sanitized += '>'
        # Print the children (contents) of the element.
        for child in elem.childNodes:
            sanitized += sanitize_node(child)
        # Print the closing tag if necessary.
        if not is_empty_element(elem):
            sanitized += '</' + htmlize(elem.name.lower()) + '>'
    # We don't do anything for unsafe elements. Sorry!
    return sanitized

try:
    # Parse a DOM tree from stdin.
    document = HTMLParser().parse(stdin, encoding='UTF-8')
    # Filter and print to stdout.
    safe_html = sanitize_node(document)
    print safe_html.encode('UTF-8', 'replace')
except Exception:
    pass
